# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import os

from django.conf import settings
from django.test import TestCase

from ..logic import process_template, list_unnecessary_loads
from ..utils import get_djangotemplates_engines, get_app


class TestLogic(TestCase):

    def test_process_template(self):
        templates_dir = settings.TEMPLATES[0]['DIRS'][0]
        dt_engines = get_djangotemplates_engines()
        app = get_app('app')
        app_path = os.path.join(app.path, 'templates')

        master_template = os.path.join(templates_dir, 'master.html')
        status = process_template(master_template, dt_engines[0].engine)
        self.assertFalse(status)

        tag_template = os.path.join(
            app_path, 'app', 'tags', 'tag_template.html')
        status = process_template(tag_template, dt_engines[0].engine)
        self.assertFalse(status)

        double_loads = os.path.join(
            app_path, 'app', 'templates', 'double_loads.html')
        status = process_template(double_loads, dt_engines[0].engine)
        self.assertTrue(status)

        double_member_load = os.path.join(
            app_path, 'app', 'templates', 'double_member_load.html')
        status = process_template(double_member_load, dt_engines[0].engine)
        self.assertTrue(status)

        from_syntax_with_tags = os.path.join(app_path, 'app', 'templates',
                                             'from_syntax_with_tags.html')
        status = process_template(from_syntax_with_tags, dt_engines[0].engine)
        self.assertFalse(status)

        from_syntax_without_tags = os.path.join(app_path, 'app',
                                                'templates',
                                                'from_syntax_without_tags.html')
        status = process_template(from_syntax_without_tags,
                                  dt_engines[0].engine)
        self.assertTrue(status)

        only_filter = os.path.join(
            app_path, 'app', 'templates', 'only_filter.html')
        status = process_template(only_filter, dt_engines[0].engine)
        self.assertFalse(status)

        with_tags = os.path.join(
            app_path, 'app', 'templates', 'with_tags.html')
        status = process_template(with_tags, dt_engines[0].engine)
        self.assertFalse(status)

        without_tags = os.path.join(
            app_path, 'app', 'templates', 'without_tags.html')
        status = process_template(without_tags, dt_engines[0].engine)
        self.assertTrue(status)

    def test_list_unnecessary_loads(self):
        status = list_unnecessary_loads()
        self.assertTrue(status)

        status = list_unnecessary_loads('app')
        self.assertTrue(status)

        status = list_unnecessary_loads('empty')
        self.assertFalse(status)

        status = list_unnecessary_loads('clean')
        self.assertFalse(status)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import os
import sys

from django import VERSION as DJANGO_VERSION
from django.apps.config import AppConfig
from django.conf import settings
from django.test import TestCase
from django.test.utils import override_settings

from ..utils import (
                     get_app,
                     get_contents,
                     get_djangotemplates_engines,
                     get_filters,
                     get_package_locations,
                     get_template_files,
                     get_templates,
                     get_templatetag_members,
                     output_as_table,
                     output_message,
                     output_template_name,
                     update_dictionary
                     )

PYTHON_VERSION = sys.version_info

if PYTHON_VERSION.major == 2:
    from StringIO import StringIO
elif PYTHON_VERSION.major == 3:
    from io import StringIO


class TestUtils(TestCase):

    def test_get_app(self):
        """
        Tests the get_app function.
        """
        # Try to fetch a non-existing app
        with self.assertRaises(LookupError):
            get_app('dummy_app')
        # Fetch the demo app and evaluate it
        app = get_app('app')
        self.assertEqual(app.label, 'app')
        self.assertIsInstance(app, AppConfig)

    def test_get_templates(self):
        templates_dir = settings.TEMPLATES[0]['DIRS'][0]
        empty_dir = templates_dir.replace('templates', 'empty_dir')
        master_template = os.path.join(templates_dir, 'master.html')
        pkg_locations = get_package_locations()
        app = get_app('app')
        app_path = os.path.join(app.path, 'templates')
        # Test templates directory
        templates = get_templates(templates_dir, pkg_locations)
        self.assertIn(master_template, templates)
        # Test empty directory
        templates = get_templates(empty_dir, pkg_locations)
        self.assertEqual(templates, [])
        # Test app directory
        templates = get_templates(app_path, pkg_locations, app)
        self.assertEqual(8, len(templates))
        self.assertIn(os.path.join(app_path, 'app', 'tags',
                                   'tag_template.html'), templates)
        self.assertIn(os.path.join(app_path, 'app', 'templates',
                                   'double_loads.html'), templates)
        self.assertIn(os.path.join(app_path, 'app', 'templates',
                                   'double_member_load.html'), templates)
        self.assertIn(os.path.join(app_path, 'app', 'templates',
                                   'from_syntax_with_tags.html'), templates)
        self.assertIn(os.path.join(app_path, 'app', 'templates',
                                   'from_syntax_without_tags.html'), templates)
        self.assertIn(os.path.join(app_path, 'app', 'templates',
                                   'only_filter.html'), templates)
        self.assertIn(os.path.join(app_path, 'app', 'templates',
                                   'with_tags.html'), templates)
        self.assertIn(os.path.join(app_path, 'app', 'templates',
                                   'without_tags.html'), templates)
        # Test external directory
        templates = get_templates(pkg_locations[0], pkg_locations)
        self.assertEqual(templates, [])

    def test_get_contents(self):
        """
        Test the get_contents function on the master.html template. Avoid
        using the DjangoTemplates class due to frequent changes between
        versions.
        """
        templates_dir = settings.TEMPLATES[0]['DIRS'][0]
        master_html = templates_dir + '/master.html'
        contents = get_contents(filepath=master_html)
        self.assertIn('DOCTYPE', contents)
        self.assertIn('html', contents)
        self.assertIn('head', contents)
        self.assertIn('title', contents)
        self.assertIn('body', contents)
        self.assertIn('{% block title %}', contents)
        self.assertIn('{% block body %}', contents)

    def test_get_djangotemplates_engines(self):
        dt_engines = get_djangotemplates_engines()
        self.assertEqual(1, len(dt_engines))
        engine = dt_engines[0]
        self.assertEqual('django', engine.name)
        self.assertTrue(engine.app_dirs)
        self.assertEqual(1, len(engine.dirs))
        self.assertTrue(engine.dirs[0].endswith('/demo/templates'))

    @override_settings(TEMPLATES=[{'BACKEND': 'foo.bar.Baz'}])
    def test_other_template_engines(self):
        dt_engines = get_djangotemplates_engines()
        self.assertEqual(0, len(dt_engines))

    def test_get_filters(self):
        token_content = '{{ somevariable|cut:"0" }}'
        filters = get_filters(token_content)
        self.assertIn('cut', filters)
        self.assertNotIn('somevariable', filters)
        self.assertNotIn('0', filters)

    def test_get_package_locations(self):
        """
        Get the directory path containing the installed 3rd party packages (in
        this instance the site-packages directory located in .tox/).
        """
        end_of_path = ('.tox/py{0}{1}-django{2}{3}/lib/'
                       'python{0}.{1}/site-packages').format(
            PYTHON_VERSION.major, PYTHON_VERSION.minor,
            DJANGO_VERSION[0], DJANGO_VERSION[1])
        pkg_locations = get_package_locations()
        for location in pkg_locations:
            self.assertTrue(location.endswith(end_of_path))

    def test_get_template_files(self):
        """
        Test the location of the master.html template file.
        """
        templates_dir = settings.TEMPLATES[0]['DIRS'][0]
        master_html = os.path.join(templates_dir, 'master.html')
        template_files = get_template_files(templates_dir)
        self.assertIn(master_html, template_files)

    def test_get_templatetag_members(self):
        output = StringIO()
        template_name = 'example.html'
        loaded_modules = {'app_tags': [1]}
        tags, filters = get_templatetag_members(template_name,
                                                loaded_modules, output)
        self.assertEqual(1, len(tags.keys()))
        self.assertIn('app_tags', tags.keys())
        self.assertIn('example_simple_tag', tags['app_tags'])
        self.assertIn('example_inclusion_tag', tags['app_tags'])
        self.assertIn('example_assignment_tag', tags['app_tags'])
        self.assertIn('app_tags', filters.keys())
        self.assertIn('plus', filters['app_tags'])
        self.assertEqual('', output.getvalue())

        # Add a non existing templatetag modules
        loaded_modules['some_lib'] = [1]
        tags, filters = get_templatetag_members(template_name,
                                                loaded_modules, output)
        self.assertEqual(output.getvalue().strip(),
                         ('Unable to locate the loaded library! '
                          'Library: some_lib; Template: example.html'))

    def test_start_output(self):
        """
        Test output to console.
        """
        output = StringIO()
        template_name = 'example.html'
        output_template_name(template_name=template_name, output=output)
        self.assertIn(template_name + '\n', output.getvalue())

    def test_output_as_table(self):
        """
        Test the output of results.

        The examples were taken from: https://pypi.python.org/pypi/tabulate
        """
        output = StringIO()
        table = [["spam", 42], ["eggs", 451], ["bacon", 0]]
        headers = ["item", "qty"]
        output_as_table(table=table, headers=headers, output=output)
        for header in headers:
            self.assertIn(header, output.getvalue())
        for row in table:
            for value in row:
                self.assertIn(str(value), output.getvalue())

    def test_output_message(self):
        output = StringIO()
        output_message(1, output)
        self.assertEqual(output.getvalue().strip(), 'No templates were found!')

        output = StringIO()
        output_message(2, output)
        self.assertEqual(output.getvalue().strip(),
                         ('Only the Django Template Engine is currently '
                          'supported!'))

        output = StringIO()
        output_message(3, output)
        self.assertEqual(output.getvalue().strip(),
                         'Your templates are clean!')

    def test_update_dictionary(self):
        dictionary = {}
        # Add new data
        dictionary = update_dictionary(dictionary=dictionary,
                                       key='module', value=1)
        self.assertIn('module', dictionary.keys())
        self.assertEqual(1, len(dictionary.keys()))
        self.assertEqual([1], dictionary.get('module'))
        # Repeat
        dictionary = update_dictionary(dictionary=dictionary,
                                       key='module', value=1)
        self.assertIn('module', dictionary.keys())
        self.assertEqual(1, len(dictionary.keys()))
        self.assertEqual([1], dictionary.get('module'))
        # Add new line number
        dictionary = update_dictionary(dictionary=dictionary,
                                       key='module', value=2)
        self.assertIn('module', dictionary.keys())
        self.assertEqual(1, len(dictionary.keys()))
        self.assertEqual([1, 2], dictionary.get('module'))



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import io
import os
import sys
from copy import deepcopy
from distutils.version import StrictVersion
from mimetypes import guess_type

from pip import get_installed_distributions
from tabulate import tabulate

from django.apps import apps
from django.conf import settings
from django.template.backends.django import DjangoTemplates
from django.template.base import InvalidTemplateLibrary

from .settings import DJANGO_VERSION

if StrictVersion(DJANGO_VERSION) > StrictVersion('1.8'):
    from django.template.base import get_library


def get_app(app_label):
    """
    Get the app object.

    :app_label: the app's label submitted by the user
    :returns: AppConfig
    """
    try:
        app = apps.get_app_config(app_label)
    except LookupError as le:
        raise le

    return app


def get_templates(directory, pkg_locations, app=None):
    """
    Traverse the project's template directories and get the paths of template
    files.

    :directory: String; path to directory
    :pkg_locations: a list of paths of 3rd party packages
    :app: AppConfig object

    :reeturns: a list of paths to template files
    """
    templates = []
    within_project = True
    for location in pkg_locations:
        if directory.startswith(location):
            within_project = False
            break
    # Get the template files from the directory
    if within_project:
        # Only one app needs to be scanned
        if app:
            if directory.startswith(app.path):
                templates = get_template_files(directory)
        else:
            templates += get_template_files(directory)

    return templates


def get_contents(filepath, encoding='UTF-8'):
    """
    Read the contents of the template file.

    A modified version of Django's get_contents method (Loader class)
    https://github.com/django/django/blob/master/django/template/loaders/filesystem.py#L22

    :filepath: absolute path to the template file
    :encoding: a string representing the engine's encoding
    :returns: contents of the template as a string
    """
    with io.open(filepath, encoding=encoding) as fp:
        return fp.read()


def get_djangotemplates_engines():
    """
    Create template engines from the parameters in the settings file.

    :returns: a list of DjangoTemplates instances
    """
    engines = []
    template_settings = settings.TEMPLATES

    for params in template_settings:
        copied_params = deepcopy(params)
        backend = copied_params.pop('BACKEND')

        if backend == 'django.template.backends.django.DjangoTemplates':
            copied_params['NAME'] = 'django'
            engines.append(DjangoTemplates(params=copied_params))
        else:
            output_message(reason=2)

    return engines


def get_filters(content):
    """
    Get filter names from the token's content.

    WARNING: Multiple filters can be used simultaneously, e.g.:
        {{ some_list|safeseq|join:", " }}

    :content: String; the token's content
    :returns: a list of filter names
    """
    filters = []
    split_content = content.split('|')

    for item in split_content[1:]:
        if ':' in item:
            item = item[:item.index(':')]
        filters.append(item)

    return filters


def get_package_locations():
    """
    Get the paths of directories where 3rd packages are installed.

    :returns: a list of absolute paths
    """
    pkg_locations = []
    installed_pkgs = get_installed_distributions(local_only=True,
                                                 include_editables=False)
    for pkg in installed_pkgs:
        if pkg.location not in pkg_locations:
            pkg_locations.append(pkg.location)

    return pkg_locations


def get_template_files(template_dir):
    """
    Scan the provided template directory and its subdirectories for template
    files.

    :returns: a list of absolute paths to template files
    """
    templates = []
    for dirpath, dirnames, filenames in os.walk(template_dir):
        for filename in filenames:
            filetype = guess_type(filename)
            maintype, subtype = filetype[0].split('/')
            if maintype == 'text':
                templates.append(os.path.join(dirpath, filename))

    return templates


def get_templatetag_members(template_name, loaded_modules, output=sys.stdout):
    """
    Get the names of tags and filters from available templatetags modules.

    :returns: {'somelib': [tags]}, {'somelib': [filters]}
    """
    tags = {}
    filters = {}
    for module in loaded_modules:
        try:
            lib = get_library(module)
        except InvalidTemplateLibrary:
            msg = ('Unable to locate the loaded library! Library: {}; '
                   'Template: {}\n').format(module, template_name)
            output.write(msg)
            tags[module] = []
            continue
        tags[module] = lib.tags.keys()
        filters[module] = lib.filters.keys()

    return tags, filters


def output_template_name(template_name, output=sys.stdout):
    """
    Output the template's name.

    :template_name: String
    :output: output destination (console=sys.stdout; testing=StringIO)
    """
    output.write('-' * len(template_name) + '\n')
    output.write(template_name + '\n')


def output_as_table(table, headers, output=sys.stdout, tablefmt='psql'):
    """
    Outputs the results in the desired format

    :table: a list of lists
    :headers: a list of strings
    :output: output destination (console=sys.stdout; testing=StringIO)
    :tablefmt: String (specific to the tabulate library)
    """
    output.write(tabulate(table, headers, tablefmt=tablefmt) + '\n')


def output_message(reason, output=sys.stdout):
    """
    Output a message to the console.

    :reason: Integer (see the dictionary below)
    :output: output destination (console=sys.stdout; testing=StringIO)
    """
    reasons = {
        1: 'No templates were found!',
        2: 'Only the Django Template Engine is currently supported!',
        3: 'Your templates are clean!'
    }
    output.write(reasons[reason] + '\n')


def update_dictionary(dictionary, key, value):
    """
    Add the key-value pair to the dictionary and return the dictionary.

    :dictionary: dict object
    :key: String (e.g. module name)
    :value: Integer (e.g. line number)
    :returns: dict object
    """
    if key not in dictionary:
        dictionary[key] = [value]
    else:
        # The same key can appear on multiple lines
        if value not in dictionary[key]:
            dictionary[key].append(value)

    return dictionary



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import sys

from .base import Template
from .utils import (
                    get_app,
                    get_contents,
                    get_package_locations,
                    get_djangotemplates_engines,
                    output_as_table,
                    output_template_name,
                    output_message,
                    get_templates
                    )


def list_unnecessary_loads(app_label=None):
    """
    Scan the project directory tree for template files and process each and
    every one of them.

    :app_label: String; app label supplied by the user

    :returns: None (outputs to the console)
    """
    if app_label:
        app = get_app(app_label)
    else:
        app = None

    dt_engines = get_djangotemplates_engines()

    for dt_engine in dt_engines:
        has_issues = False
        templates = []
        # Get the locations of installed packages
        pkg_locations = get_package_locations()
        # Get template directories located within the project
        for directory in dt_engine.template_dirs:
            templates += get_templates(directory, pkg_locations, app)

        if templates:
            for template in templates:
                status = process_template(template, dt_engine.engine)
                if status:
                    has_issues = status
            if not has_issues:
                output_message(reason=3)
        else:
            output_message(reason=1)

    return has_issues


def process_template(filepath, engine):
    """
    Process the specified template

    :filepath: String; the absolute path to the template file
    :engine: Engine object

    :returns: Boolean (does the template file have issues or not)
    """
    has_issues = False
    # Get the template's contents
    source = get_contents(filepath=filepath,
                          encoding=engine.file_charset)
    # Create and process the template
    template = Template(template_string=source, engine=engine, name=filepath)
    # Prepare output
    duplicate_table, duplicate_headers = template.list_duplicates()
    unutilized_table, unutilized_headers = template.list_unutilized_items()
    if duplicate_table or unutilized_table:
        output_template_name(template_name=template.name, output=sys.stdout)
        has_issues = True
    # Display the table that contains duplicate loads
    if duplicate_table:
        output_as_table(table=duplicate_table,
                        headers=duplicate_headers, output=sys.stdout)
    # Display the table that contains unutilized loads
    if unutilized_table:
        output_as_table(table=unutilized_table,
                        headers=unutilized_headers, output=sys.stdout)

    return has_issues



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django import template

register = template.Library()


@register.inclusion_tag('app/tags/tag_template.html')
def example_inclusion_tag():
    """
    An example of an inclusion tag.
    """
    return {'tag_name': 'example_inclusion_tag'}


@register.simple_tag
def example_simple_tag(*args):
    """
    An example of a simple tag.
    """
    return 'example_simple_tag'


@register.assignment_tag
def example_assignment_tag():
    """
    An example of an assignment tag.

    From the Django documentation:
        Deprecated since version 1.9:
        simple_tag can now store results in a template variable and should be
        used instead.
    """
    return 'example_assignment_tag'


@register.filter(name='plus')
def addition(value, arg=1):
    return value + arg



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from distutils.version import StrictVersion

from django.core.management.base import BaseCommand
from django.utils.translation import ugettext_lazy as _

from ...logic import list_unnecessary_loads
from ...settings import DJANGO_VERSION

if StrictVersion(DJANGO_VERSION) < StrictVersion('1.8'):
    raise NotImplementedError('Minimal supported version of Django: 1.8')
elif StrictVersion(DJANGO_VERSION) > StrictVersion('1.9'):
    raise NotImplementedError('Django version {} is not yet supported'.format(
                                DJANGO_VERSION))


class Command(BaseCommand):
    help = 'List unutilized templatetag libraries'

    def add_arguments(self, parser):
        parser.add_argument(
            '-a', '--app', nargs='?', type=str, action='store', required=False,
            help=_('The label of the application that needs to be scanned'))

    def handle(self, *args, **options):
        # Find the app
        app_label = options.get('app', None)
        self.stdout.write(
            'Has issues: {}'.format(str(list_unnecessary_loads(app_label))))



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.template.base import Lexer, Template as BaseTemplate

from .settings import BUILT_IN_TAGS, BUILT_IN_TAG_VALUES, BUILT_IN_FILTERS
from .utils import get_filters, get_templatetag_members, update_dictionary


class Template(BaseTemplate):
    """
    An override of Django's Template class.

    After calling the parent class, the template is analyzed for duplicates
    and unnecessary loads.

    Additional attributes:
    :tokens: a list of tokens found in the template
    :loaded_modules: a dictionary of loaded modules
    :loaded_members: a dictionary of loaded tags/filters
    :used_tags: a list of custom tags used in the template
    :used_filters: a list of custom filters used in the template
    :tags: a dictionary of custom tags loaded into the template
    :filters: a dictionary of custom filters loaded into the template
    :utilized_modules: a dictionary of utilization statuses
    :utilized_members: a dictionary of utilization statuses
    """

    def __init__(self, template_string, origin=None, name=None, engine=None):
        super(Template, self).__init__(template_string, origin, name, engine)

        # Used for backwards compatibility (implemented in Django 1.9)
        if not hasattr(self, 'source'):
            self.source = template_string

        self.tokens = self._get_tokens()
        # The manually specified (loaded) modules and members (tags/filters)
        self.loaded_modules, self.loaded_members = self._parse_load_block()
        self.used_tags = self._get_used_tags()
        self.used_filters = self._get_used_filters()
        # Get the tags and filters available to this template
        self.tags, self.filters = get_templatetag_members(
            self.name, self.loaded_modules)
        # Find utilized modules, tags and filters
        self.utilized_modules = self._get_utilized_modules()
        self.utilized_members = self._get_utilized_members()

    def list_duplicates(self):
        """
        List duplicate results, i.e. duplicate library, tag or filter loads.
        If possible, tries to combine the results in the same row.

        :returns: table (list of lists), header (list of header names)
        """

        temp_table = {}
        # Find duplicate library loads
        for module in self.loaded_modules:
            lines = self.loaded_modules[module]
            lines_str = ', '.join(map(str, lines))
            if len(lines) > 1 and lines_str not in temp_table.keys():
                temp_table[lines_str] = [module, []]

        # Find duplicate member loads
        for member in self.loaded_members:
            lines = self.loaded_members[member]
            lines_str = ', '.join(map(str, lines))

            if len(lines) > 1:
                temp_table[lines_str][1].append(member)

        for key in temp_table:
            if temp_table[key][1] == []:
                temp_table[key][1] = None
            else:
                temp_table[key][1] = '; '.join(temp_table[key][1])

        # Prepare output format
        headers = ['Duplicate module', 'Duplicate tag/filter', 'Line number']
        table = []
        if temp_table:
            for key in temp_table:
                row = temp_table[key]
                row.append(key)
                table.append(row)

        return table, headers

    def list_unutilized_items(self):
        """
        List unutilized modules, tags and filters in a single table
        """
        modules = []
        members = []
        # List unutilized modules
        if self.utilized_modules:
            for module in self.utilized_modules:
                if not self.utilized_modules[module]:
                    modules.append(module)

        # List unutilized tags/filters
        if self.utilized_members:
            for member in self.utilized_members:
                if not self.utilized_members[member]:
                    members.append(member)

        if len(modules) > len(members):
            diff = len(modules) - len(members)
            members += [None] * diff
        else:
            diff = len(members) - len(modules)
            modules += [None] * diff

        headers = ['Unutilized module', 'Unutilized tag/filter']
        table = zip(modules, members)

        return list(table), headers

    def _get_utilized_members(self):
        """
        Separates the loaded tags based on their utilization.

        :returns: {'tag_name': Boolean}
        """
        utilized_members = {}

        for member in self.loaded_members:
            utilized = False
            if member in self.used_tags or member in self.used_filters:
                utilized = True
            utilized_members[member] = utilized

        return utilized_members

    def _get_utilized_modules(self):
        """
        Separates the loaded modules based on their utilization.

        :returns: {'module_name': Boolean}
        """
        utilized_modules = {}

        for module in self.loaded_modules:
            utilized = False

            for tag in self.used_tags:
                if tag in self.tags[module]:
                    utilized = True
                    break

            if not utilized and self.used_filters:
                for custom_filter in self.used_filters:
                    if custom_filter in self.filters[module]:
                        utilized = True
                        break

            utilized_modules[module] = utilized

        return utilized_modules

    def _get_tokens(self):
        """
        Get the list of tokens from the template source.

        A modified version of Django's compile_nodelist method (Template class)
        https://github.com/django/django/blob/master/django/template/base.py#L221

        :returns: a list of Tokens
        """
        # From Django's source code
        if self.engine.debug:
            from django.template.debug import DebugLexer
            lexer = DebugLexer(self.source, self.origin)
        else:
            lexer = Lexer(self.source, self.origin)

        return lexer.tokenize()

    def _parse_load_block(self):
        """
        Get the names of loaded templatetags modules as well as individually
        loaded tags and the line numbers they are located at.

        :returns: {'module': [line_numbers]}, {'member': [line_numbers]}
        """

        modules = {}
        members = {}

        for token in self.tokens:
            token_content = token.split_contents()
            # Extract load blocks
            if token.token_type == 2 and token_content[0] == 'load':
                # FROM syntax is used; individual members are loaded
                if len(token_content) >= 4 and token_content[-2] == 'from':
                    # Add loaded module
                    module = token_content[-1]
                    modules = update_dictionary(modules, module, token.lineno)
                    # Add loaded members
                    loaded_members = token_content[1:-2]
                    for member in loaded_members:
                        members = update_dictionary(members, member, token.lineno)
                # Regular syntax
                else:
                    # Multiple modules can be imported in the same load block
                    templatetags_modules = token_content[1:]
                    for module in templatetags_modules:
                        modules = update_dictionary(modules, module, token.lineno)

        return modules, members

    def _get_used_tags(self):
        """
        Get the list of custom template tags used in the template.

        :returns: a list of custom tag names
        """
        used_tags = []
        for token in self.tokens:
            token_content = token.split_contents()
            # Extract blocks that do not contain one of the built-in tags
            if (token.token_type == 2 and
                    token_content[0] not in BUILT_IN_TAGS.keys() and
                    # Skip built-in 'end' tags
                    token_content[0] not in BUILT_IN_TAG_VALUES):
                # Extract only the name of the template tag (ignore arguments)
                used_tags.append(token_content[0])

        return used_tags

    def _get_used_filters(self):
        """
        Get the list of custom filters used in the template.

        :returns: a list of custom filter names
        """
        used_filters = []

        for token in self.tokens:
            filters = []
            token_content = token.split_contents()

            # Variable token
            if token.token_type == 1 and '|' in token_content[0]:
                filters += get_filters(token_content[0])

            # Tag token
            elif token.token_type == 2:
                if '|' in ' '.join(token_content):
                    for item in token_content:
                        if '|' in item:
                            filters += get_filters(item)

            # Exclude built-in filters
            for filter_name in filters:
                if (filter_name not in BUILT_IN_FILTERS and
                        filter_name not in used_filters):
                    used_filters.append(filter_name)

        return used_filters



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import os

from django.apps import apps
from django.conf import settings
from django.test import TestCase
from django.test.utils import override_settings

from ..base import Template
from ..utils import get_contents


class TestBase(TestCase):

    @classmethod
    def setUpClass(cls):
        """
        The setUpClass is used because the templates paths are not changed in
        the following tests.
        """
        super(TestBase, cls).setUpClass()
        # Find directories with templates
        templates_dir = settings.TEMPLATES[0]['DIRS'][0]
        app = apps.get_app_config('app')
        app_templates = os.path.join(app.path, 'templates', 'app', 'templates')
        app_tags = os.path.join(app.path, 'templates', 'app', 'tags')
        # Save the path of each template for further testing
        cls.master_template = os.path.join(templates_dir, 'master.html')
        cls.tag_template = os.path.join(app_tags, 'tag_template.html')
        cls.double_loads = os.path.join(app_templates, 'double_loads.html')
        cls.with_tags = os.path.join(app_templates, 'with_tags.html')
        cls.without_tags = os.path.join(app_templates, 'without_tags.html')
        cls.from_syntax_with_tags = os.path.join(app_templates,
                                                 'from_syntax_with_tags.html')
        cls.from_syntax_without_tags = os.path.join(app_templates,
                                                    'from_syntax_without_tags.html')
        cls.double_member_load = os.path.join(app_templates,
                                              'double_member_load.html')
        cls.only_filter = os.path.join(app_templates, 'only_filter.html')

    def test_get_tokens_master_template(self):
        master_template = Template(
            template_string=get_contents(self.master_template),
            name=self.master_template)

        self.assertEqual(master_template.tokens[1].split_contents(),
                         ['block', 'title'])
        self.assertEqual(master_template.tokens[3].split_contents(),
                         ['endblock', 'title'])
        self.assertEqual(master_template.tokens[5].split_contents(),
                         ['block', 'body'])
        self.assertEqual(master_template.tokens[7].split_contents(),
                         ['endblock', 'body'])

    @override_settings(DEBUG=True)
    def test_get_tokens_master_template_debug_mode(self):
        """
        The settings override is tested only once because the code using the
        DebugLexer is extracted from Django's source code.
        """
        master_template = Template(
            template_string=get_contents(self.master_template),
            name=self.master_template)

        self.assertEqual(master_template.tokens[1].split_contents(),
                         ['block', 'title'])
        self.assertEqual(master_template.tokens[3].split_contents(),
                         ['endblock', 'title'])
        self.assertEqual(master_template.tokens[5].split_contents(),
                         ['block', 'body'])
        self.assertEqual(master_template.tokens[7].split_contents(),
                         ['endblock', 'body'])

    def test_get_tokens_tag_template(self):
        # Test tag template
        tag_template = Template(
            template_string=get_contents(self.tag_template),
            name=self.tag_template)
        self.assertEqual(tag_template.tokens[1].contents, 'tag_name')

    def test_get_tokens_double_loads(self):
        # Test double loads
        double_loads = Template(
            template_string=get_contents(self.double_loads),
            name=self.double_loads)

        self.assertEqual(double_loads.tokens[0].split_contents(),
                         ['extends', '"master.html"'])
        self.assertEqual(double_loads.tokens[2].split_contents(),
                         ['load', 'app_tags'])
        self.assertEqual(double_loads.tokens[4].split_contents(),
                         ['load', 'app_tags'])
        self.assertEqual(double_loads.tokens[6].split_contents(),
                         ['block', 'body'])
        self.assertEqual(double_loads.tokens[8].split_contents(),
                         ['endblock', 'body'])

    def test_get_tokens_with_tags(self):
        with_tags = Template(
            template_string=get_contents(self.with_tags),
            name=self.with_tags)
        self.assertEqual(with_tags.tokens[0].split_contents(),
                         ['extends', '"master.html"'])
        self.assertEqual(with_tags.tokens[2].split_contents(),
                         ['load', 'app_tags'])
        self.assertEqual(with_tags.tokens[4].split_contents(),
                         ['block', 'body'])
        self.assertEqual(with_tags.tokens[6].split_contents(),
                         ['example_inclusion_tag'])
        self.assertEqual(with_tags.tokens[8].split_contents(),
                         ['example_simple_tag'])
        self.assertEqual(with_tags.tokens[10].split_contents(),
                         ['example_assignment_tag', 'as', 'example'])
        self.assertEqual(with_tags.tokens[12].split_contents(),
                         ['example_simple_tag', '2|plus:5'])
        self.assertEqual(with_tags.tokens[14].split_contents(),
                         ['endblock', 'body'])

    def test_get_tokens_from_syntax_with_tags(self):
        from_syntax_with_tags = Template(
            template_string=get_contents(self.from_syntax_with_tags),
            name=self.from_syntax_with_tags)

        self.assertEqual(from_syntax_with_tags.tokens[0].split_contents(),
                         ['extends', '"master.html"'])
        self.assertEqual(from_syntax_with_tags.tokens[2].split_contents(),
                         ['load', 'example_inclusion_tag',
                          'example_simple_tag', 'example_assignment_tag',
                          'plus', 'from', 'app_tags'])
        self.assertEqual(from_syntax_with_tags.tokens[4].split_contents(),
                         ['block', 'body'])
        self.assertEqual(from_syntax_with_tags.tokens[6].split_contents(),
                         ['example_inclusion_tag'])
        self.assertEqual(from_syntax_with_tags.tokens[8].split_contents(),
                         ['example_simple_tag'])
        self.assertEqual(from_syntax_with_tags.tokens[10].split_contents(),
                         ['example_assignment_tag', 'as', 'example'])
        self.assertEqual(from_syntax_with_tags.tokens[12].split_contents(),
                         ['2|plus:5'])
        self.assertEqual(from_syntax_with_tags.tokens[14].split_contents(),
                         ['endblock', 'body'])

    def test_get_tokens_without_tags(self):
        without_tags = Template(
            template_string=get_contents(self.without_tags),
            name=self.without_tags)

        self.assertEqual(without_tags.tokens[0].split_contents(),
                         ['extends', '"master.html"'])
        self.assertEqual(without_tags.tokens[2].split_contents(),
                         ['load', 'app_tags'])
        self.assertEqual(without_tags.tokens[4].split_contents(),
                         ['block', 'body'])
        self.assertEqual(without_tags.tokens[6].split_contents(),
                         ['endblock', 'body'])

    def test_get_tokens_from_syntax_without_tags(self):
        # Test double loads
        from_syntax_without_tags = Template(
            template_string=get_contents(self.from_syntax_without_tags),
            name=self.from_syntax_without_tags)

        self.assertEqual(from_syntax_without_tags.tokens[0].split_contents(),
                         ['extends', '"master.html"'])
        self.assertEqual(from_syntax_without_tags.tokens[2].split_contents(),
                         ['load', 'example_simple_tag', 'plus', 'from',
                          'app_tags'])
        self.assertEqual(from_syntax_without_tags.tokens[4].split_contents(),
                         ['block', 'body'])
        self.assertEqual(from_syntax_without_tags.tokens[6].split_contents(),
                         ['endblock', 'body'])

    def test_get_tokens_double_member_load(self):
        double_member_load = Template(
            template_string=get_contents(self.double_member_load),
            name=self.double_member_load)
        self.assertEqual(double_member_load.tokens[0].split_contents(),
                         ['extends', '"master.html"'])
        self.assertEqual(double_member_load.tokens[2].split_contents(),
                         ['load', 'example_simple_tag', 'plus', 'from',
                          'app_tags'])
        self.assertEqual(double_member_load.tokens[4].split_contents(),
                         ['load', 'example_simple_tag', 'plus', 'from',
                          'app_tags'])
        self.assertEqual(double_member_load.tokens[6].split_contents(),
                         ['block', 'body'])
        self.assertEqual(double_member_load.tokens[8].split_contents(),
                         ['endblock', 'body'])

    def test_get_tokens_only_filter(self):
        only_filter = Template(
            template_string=get_contents(self.only_filter),
            name=self.only_filter)
        self.assertEqual(only_filter.tokens[0].split_contents(),
                         ['extends', '"master.html"'])
        self.assertEqual(only_filter.tokens[2].split_contents(),
                         ['load', 'plus', 'from', 'app_tags'])
        self.assertEqual(only_filter.tokens[4].split_contents(),
                         ['block', 'body'])
        self.assertEqual(only_filter.tokens[6].split_contents(),
                         ['2|plus:5'])
        self.assertEqual(only_filter.tokens[8].split_contents(),
                         ['endblock', 'body'])

    def test_parse_load_block(self):

        master_template = Template(
            template_string=get_contents(self.master_template),
            name=self.master_template)
        self.assertEqual(master_template.loaded_modules, {})
        self.assertEqual(master_template.loaded_members, {})

        tag_template = Template(
            template_string=get_contents(self.tag_template),
            name=self.tag_template)
        self.assertEqual(tag_template.loaded_modules, {})
        self.assertEqual(tag_template.loaded_members, {})

        double_loads = Template(
            template_string=get_contents(self.double_loads),
            name=self.double_loads)
        self.assertEqual(double_loads.loaded_modules,
                         {'app_tags': [2, 3]})
        self.assertEqual(double_loads.loaded_members, {})

        with_tags = Template(
            template_string=get_contents(self.with_tags),
            name=self.with_tags)
        self.assertEqual(with_tags.loaded_modules,
                         {'app_tags': [2]})
        self.assertEqual(with_tags.loaded_members, {})

        from_syntax_with_tags = Template(
            template_string=get_contents(self.from_syntax_with_tags),
            name=self.from_syntax_with_tags)
        self.assertEqual(from_syntax_with_tags.loaded_modules,
                         {'app_tags': [2]})
        self.assertEqual(from_syntax_with_tags.loaded_members,
                         {'example_assignment_tag': [2],
                          'example_inclusion_tag': [2],
                          'example_simple_tag': [2],
                          'plus': [2]})

        without_tags = Template(
            template_string=get_contents(self.without_tags),
            name=self.without_tags)
        self.assertEqual(without_tags.loaded_modules,
                         {'app_tags': [2]})
        self.assertEqual(without_tags.loaded_members, {})

        from_syntax_without_tags = Template(
            template_string=get_contents(self.from_syntax_without_tags),
            name=self.from_syntax_without_tags)
        self.assertEqual(from_syntax_without_tags.loaded_modules,
                         {'app_tags': [2]})
        self.assertEqual(from_syntax_without_tags.loaded_members,
                         {'example_simple_tag': [2],
                          'plus': [2]})

        double_member_load = Template(
            template_string=get_contents(self.double_member_load),
            name=self.double_member_load)
        self.assertEqual(double_member_load.loaded_modules,
                         {'app_tags': [2, 3]})
        self.assertEqual(double_member_load.loaded_members,
                         {'example_simple_tag': [2, 3],
                          'plus': [2, 3]})

        only_filter = Template(
            template_string=get_contents(self.only_filter),
            name=self.only_filter)
        self.assertEqual(only_filter.loaded_modules,
                         {'app_tags': [2]})
        self.assertEqual(only_filter.loaded_members, {'plus': [2]})

    def test_get_used_tags_and_filters(self):

        master_template = Template(
            template_string=get_contents(self.master_template),
            name=self.master_template)
        self.assertEqual(master_template.used_tags, [])
        self.assertEqual(master_template.used_filters, [])

        tag_template = Template(
            template_string=get_contents(self.tag_template),
            name=self.tag_template)
        self.assertEqual(tag_template.used_tags, [])
        self.assertEqual(tag_template.used_filters, [])

        double_loads = Template(
            template_string=get_contents(self.double_loads),
            name=self.double_loads)
        self.assertEqual(double_loads.used_tags, [])
        self.assertEqual(double_loads.used_filters, [])

        with_tags = Template(
            template_string=get_contents(self.with_tags),
            name=self.with_tags)
        self.assertEqual(with_tags.used_tags,
                         ['example_inclusion_tag', 'example_simple_tag',
                          'example_assignment_tag', 'example_simple_tag'])
        self.assertEqual(with_tags.used_filters, ['plus'])

        from_syntax_with_tags = Template(
            template_string=get_contents(self.from_syntax_with_tags),
            name=self.from_syntax_with_tags)
        self.assertEqual(from_syntax_with_tags.used_tags,
                         ['example_inclusion_tag', 'example_simple_tag',
                          'example_assignment_tag'])
        self.assertEqual(from_syntax_with_tags.used_filters, ['plus'])

        without_tags = Template(
            template_string=get_contents(self.without_tags),
            name=self.without_tags)
        self.assertEqual(without_tags.used_tags, [])
        self.assertEqual(without_tags.used_filters, [])

        from_syntax_without_tags = Template(
            template_string=get_contents(self.from_syntax_without_tags),
            name=self.from_syntax_without_tags)
        self.assertEqual(from_syntax_without_tags.used_tags, [])
        self.assertEqual(from_syntax_without_tags.used_filters, [])

        double_member_load = Template(
            template_string=get_contents(self.double_member_load),
            name=self.double_member_load)
        self.assertEqual(double_member_load.used_tags, [])
        self.assertEqual(double_member_load.used_filters, [])

        only_filter = Template(
            template_string=get_contents(self.only_filter),
            name=self.only_filter)
        self.assertEqual(only_filter.used_tags, [])
        self.assertEqual(only_filter.used_filters, ['plus'])

    def test_get_templatetags_members(self):

        master_template = Template(
            template_string=get_contents(self.master_template),
            name=self.master_template)
        self.assertEqual(master_template.tags, {})
        self.assertEqual(master_template.filters, {})

        tag_template = Template(
            template_string=get_contents(self.tag_template),
            name=self.tag_template)
        self.assertEqual(tag_template.tags, {})
        self.assertEqual(tag_template.filters, {})

        double_loads = Template(
            template_string=get_contents(self.double_loads),
            name=self.double_loads)

        self.assertIn('app_tags', double_loads.tags.keys())
        self.assertIn('example_inclusion_tag', double_loads.tags['app_tags'])
        self.assertIn('example_simple_tag', double_loads.tags['app_tags'])
        self.assertIn('example_assignment_tag', double_loads.tags['app_tags'])
        self.assertIn('app_tags', double_loads.filters.keys())
        self.assertIn('plus', double_loads.filters['app_tags'])

        with_tags = Template(
            template_string=get_contents(self.with_tags),
            name=self.with_tags)
        self.assertIn('app_tags', with_tags.tags.keys())
        self.assertIn('example_inclusion_tag', with_tags.tags['app_tags'])
        self.assertIn('example_simple_tag', with_tags.tags['app_tags'])
        self.assertIn('example_assignment_tag', with_tags.tags['app_tags'])
        self.assertIn('app_tags', with_tags.filters.keys())
        self.assertIn('plus', with_tags.filters['app_tags'])

        from_syntax_with_tags = Template(
            template_string=get_contents(self.from_syntax_with_tags),
            name=self.from_syntax_with_tags)
        self.assertIn('app_tags', from_syntax_with_tags.tags.keys())
        self.assertIn('example_inclusion_tag',
                      from_syntax_with_tags.tags['app_tags'])
        self.assertIn('example_simple_tag',
                      from_syntax_with_tags.tags['app_tags'])
        self.assertIn('example_assignment_tag',
                      from_syntax_with_tags.tags['app_tags'])
        self.assertIn('app_tags', from_syntax_with_tags.filters.keys())
        self.assertIn('plus', from_syntax_with_tags.filters['app_tags'])

        without_tags = Template(
            template_string=get_contents(self.without_tags),
            name=self.without_tags)
        self.assertIn('app_tags', without_tags.tags.keys())
        self.assertIn('example_inclusion_tag', without_tags.tags['app_tags'])
        self.assertIn('example_simple_tag', without_tags.tags['app_tags'])
        self.assertIn('example_assignment_tag', without_tags.tags['app_tags'])
        self.assertIn('app_tags', without_tags.filters.keys())
        self.assertIn('plus', without_tags.filters['app_tags'])

        from_syntax_without_tags = Template(
            template_string=get_contents(self.from_syntax_without_tags),
            name=self.from_syntax_without_tags)
        self.assertIn('app_tags', from_syntax_without_tags.tags.keys())
        self.assertIn('example_simple_tag',
                      from_syntax_without_tags.tags['app_tags'])
        self.assertIn('app_tags', from_syntax_without_tags.filters.keys())
        self.assertIn('plus', from_syntax_without_tags.filters['app_tags'])

        double_member_load = Template(
            template_string=get_contents(self.double_member_load),
            name=self.double_member_load)
        self.assertIn('app_tags', double_member_load.tags.keys())
        self.assertIn('example_simple_tag',
                      double_member_load.tags['app_tags'])
        self.assertIn('app_tags', double_member_load.filters.keys())
        self.assertIn('plus', double_member_load.filters['app_tags'])

        only_filter = Template(
            template_string=get_contents(self.only_filter),
            name=self.only_filter)
        self.assertEqual(master_template.tags, {})
        self.assertIn('app_tags', only_filter.filters.keys())
        self.assertIn('plus', only_filter.filters['app_tags'])

    def test_get_utilized_modules_and_members(self):
        master_template = Template(
            template_string=get_contents(self.master_template),
            name=self.master_template)
        self.assertEqual(master_template.utilized_modules, {})
        self.assertEqual(master_template.utilized_members, {})

        tag_template = Template(
            template_string=get_contents(self.tag_template),
            name=self.tag_template)
        self.assertEqual(tag_template.utilized_modules, {})
        self.assertEqual(tag_template.utilized_members, {})

        double_loads = Template(
            template_string=get_contents(self.double_loads),
            name=self.double_loads)
        self.assertFalse(double_loads.utilized_modules['app_tags'])
        self.assertEqual(double_loads.utilized_members, {})

        with_tags = Template(
            template_string=get_contents(self.with_tags),
            name=self.with_tags)
        self.assertTrue(with_tags.utilized_modules['app_tags'])
        self.assertEqual(with_tags.utilized_members, {})

        from_syntax_with_tags = Template(
            template_string=get_contents(self.from_syntax_with_tags),
            name=self.from_syntax_with_tags)
        self.assertTrue(from_syntax_with_tags.utilized_modules['app_tags'])
        self.assertTrue(
            from_syntax_with_tags.utilized_members['example_simple_tag'])
        self.assertTrue(
            from_syntax_with_tags.utilized_members['example_inclusion_tag'])
        self.assertTrue(
            from_syntax_with_tags.utilized_members['example_assignment_tag'])
        self.assertTrue(
            from_syntax_with_tags.utilized_members['plus'])

        without_tags = Template(
            template_string=get_contents(self.without_tags),
            name=self.without_tags)
        self.assertFalse(without_tags.utilized_modules['app_tags'])
        self.assertEqual(without_tags.utilized_members, {})

        from_syntax_without_tags = Template(
            template_string=get_contents(self.from_syntax_without_tags),
            name=self.from_syntax_without_tags)

        self.assertFalse(from_syntax_without_tags.utilized_modules['app_tags'])
        self.assertFalse(
            from_syntax_without_tags.utilized_members['example_simple_tag'])
        self.assertFalse(from_syntax_without_tags.utilized_members['plus'])

        double_member_load = Template(
            template_string=get_contents(self.double_member_load),
            name=self.double_member_load)
        self.assertFalse(double_member_load.utilized_modules['app_tags'])
        self.assertFalse(double_member_load.utilized_members['plus'])

        only_filter = Template(
            template_string=get_contents(self.only_filter),
            name=self.only_filter)
        self.assertTrue(only_filter.utilized_modules['app_tags'])
        self.assertTrue(only_filter.utilized_members['plus'])

    def test_list_duplicates(self):
        master_template = Template(
            template_string=get_contents(self.master_template),
            name=self.master_template)
        self.assertEqual(master_template.list_duplicates(),
                         ([], ['Duplicate module', 'Duplicate tag/filter',
                               'Line number']))

        tag_template = Template(
            template_string=get_contents(self.tag_template),
            name=self.tag_template)
        self.assertEqual(tag_template.list_duplicates(),
                         ([], ['Duplicate module', 'Duplicate tag/filter',
                               'Line number']))

        double_loads = Template(
            template_string=get_contents(self.double_loads),
            name=self.double_loads)
        self.assertEqual(double_loads.list_duplicates(),
                         ([['app_tags', None, '2, 3']],
                          ['Duplicate module', 'Duplicate tag/filter',
                           'Line number']))

        with_tags = Template(
            template_string=get_contents(self.with_tags),
            name=self.with_tags)
        self.assertEqual(with_tags.list_duplicates(),
                         ([], ['Duplicate module', 'Duplicate tag/filter',
                               'Line number']))

        from_syntax_with_tags = Template(
            template_string=get_contents(self.from_syntax_with_tags),
            name=self.from_syntax_with_tags)
        self.assertEqual(from_syntax_with_tags.list_duplicates(),
                         ([], ['Duplicate module', 'Duplicate tag/filter',
                               'Line number']))

        without_tags = Template(
            template_string=get_contents(self.without_tags),
            name=self.without_tags)
        self.assertEqual(without_tags.list_duplicates(),
                         ([], ['Duplicate module', 'Duplicate tag/filter',
                               'Line number']))

        from_syntax_without_tags = Template(
            template_string=get_contents(self.from_syntax_without_tags),
            name=self.from_syntax_without_tags)
        self.assertEqual(from_syntax_without_tags.list_duplicates(),
                         ([], ['Duplicate module', 'Duplicate tag/filter',
                               'Line number']))

        double_member_load = Template(
            template_string=get_contents(self.double_member_load),
            name=self.double_member_load)
        double_member_load_table,\
            double_member_load_header = double_member_load.list_duplicates()
        self.assertIn('app_tags', double_member_load_table[0])
        self.assertIn('plus', double_member_load_table[0][1])
        self.assertIn('example_simple_tag', double_member_load_table[0][1])
        self.assertIn('2, 3', double_member_load_table[0])
        self.assertEqual(double_member_load_header,
                         ['Duplicate module', 'Duplicate tag/filter',
                          'Line number'])

        only_filter = Template(
            template_string=get_contents(self.only_filter),
            name=self.only_filter)
        self.assertEqual(only_filter.list_duplicates(),
                         ([], ['Duplicate module', 'Duplicate tag/filter',
                               'Line number']))

    def test_list_unutilized_items(self):
        master_template = Template(
            template_string=get_contents(self.master_template),
            name=self.master_template)

        self.assertEqual(master_template.list_unutilized_items(),
                         ([], ['Unutilized module', 'Unutilized tag/filter']))

        tag_template = Template(
            template_string=get_contents(self.tag_template),
            name=self.tag_template)
        self.assertEqual(tag_template.list_unutilized_items(),
                         ([], ['Unutilized module', 'Unutilized tag/filter']))

        double_loads = Template(
            template_string=get_contents(self.double_loads),
            name=self.double_loads)
        self.assertEqual(double_loads.list_unutilized_items(),
                         ([('app_tags', None)],
                          ['Unutilized module', 'Unutilized tag/filter']))

        with_tags = Template(
            template_string=get_contents(self.with_tags),
            name=self.with_tags)
        self.assertEqual(with_tags.list_unutilized_items(),
                         ([], ['Unutilized module', 'Unutilized tag/filter']))

        from_syntax_with_tags = Template(
            template_string=get_contents(self.from_syntax_with_tags),
            name=self.from_syntax_with_tags)
        self.assertEqual(from_syntax_with_tags.list_unutilized_items(),
                         ([], ['Unutilized module', 'Unutilized tag/filter']))

        without_tags = Template(
            template_string=get_contents(self.without_tags),
            name=self.without_tags)
        self.assertEqual(without_tags.list_unutilized_items(),
                         ([('app_tags', None)],
                          ['Unutilized module', 'Unutilized tag/filter']))

        from_syntax_without_tags = Template(
            template_string=get_contents(self.from_syntax_without_tags),
            name=self.from_syntax_without_tags)
        table, header = from_syntax_without_tags.list_unutilized_items()
        # Merge the two rows due to unpredictable ordering of members
        all_rows = table[0] + table[1]
        self.assertEqual(
            header, ['Unutilized module', 'Unutilized tag/filter'])
        self.assertIn('app_tags', all_rows)
        self.assertIn('example_simple_tag', all_rows)
        self.assertIn('plus', all_rows)

        double_member_load = Template(
            template_string=get_contents(self.double_member_load),
            name=self.double_member_load)
        table, header = double_member_load.list_unutilized_items()
        # Merge the two rows due to unpredictable ordering of members
        all_rows = table[0] + table[1]
        self.assertEqual(
            header, ['Unutilized module', 'Unutilized tag/filter'])
        self.assertIn('app_tags', all_rows)
        self.assertIn('example_simple_tag', all_rows)
        self.assertIn('plus', all_rows)

        only_filter = Template(
            template_string=get_contents(self.only_filter),
            name=self.only_filter)
        self.assertEqual(only_filter.list_unutilized_items(),
                         ([], ['Unutilized module', 'Unutilized tag/filter']))




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import sys
from distutils.version import StrictVersion

from django import get_version

DJANGO_VERSION = get_version()

if StrictVersion(DJANGO_VERSION) > StrictVersion('1.8'):
    from django.template.defaultfilters import register as filter_lib
    from django.templatetags.cache import register as cache_lib
    from django.templatetags.future import register as future_lib
    from django.templatetags.i18n import register as i18n_lib
    from django.templatetags.l10n import register as l10n_lib
    from django.templatetags.static import register as static_lib
    from django.templatetags.tz import register as tz_lib

PYTHON_VERSION = sys.version_info

BUILT_IN_FILTERS = filter_lib.filters.keys()
CACHE_FILTERS = cache_lib.filters.keys()
FUTURE_FILTERS = future_lib.filters.keys()
I18N_FILTERS = i18n_lib.filters.keys()
L10N_FILTERS = l10n_lib.filters.keys()
STATIC_FILTERS = static_lib.filters.keys()
TZ_FILTERS = tz_lib.filters.keys()

# https://docs.djangoproject.com/en/1.9/ref/templates/builtins/#built-in-tag-reference
BUILT_IN_TAGS = {
    'as': None,
    'autoescape': 'endautoescape',
    'block': 'endblock',
    'blocktrans': 'endblocktrans',
    'comment': 'endcomment',
    'csrf_token': None,
    'cycle': None,
    'debug': None,
    'extends': None,
    'filter': 'endfilter',
    'firstof': None,
    'for': 'endfor',
    'elif': None,
    'else': None,
    'empty': None,
    'if': 'endif',
    'ifchanged': 'endifchanged',
    'ifequal': 'endifequal',
    'ifnotequal': 'endifnotequal',
    'in': None,
    'include': None,
    'load': None,
    'lorem': None,
    'not': None,
    'now': None,
    'spaceless': 'endspaceless',
    'ssi': None,
    'templatetag': None,
    'url': None,
    'verbatim': 'endverbatim',
    'widthratio': None,
    'with': 'endwith'
}

I18N_TAGS = {
    'trans': 'endtrans',
    'blocktrans': 'endblocktrans',
    'plural': None,
    'get_language_info_list': None,
    'get_available_languages': None,
    'get_language_info': None,
    'language': None,
    'get_current_language': None,
    'get_current_language_bidi': None
}

L10N_TAGS = {
    'localize': 'endlocalize'
}

CACHE_TAGS = {
    'cache': 'endcache'
}

STATIC_TAGS = {
    'static': None,
    'get_media_prefix': None,
    'get_static_prefix': None
}

if PYTHON_VERSION.major == 2:
    BUILT_IN_TAG_VALUES = set(BUILT_IN_TAGS.values() +
                              I18N_TAGS.values() +
                              L10N_TAGS.values() +
                              CACHE_TAGS.values() +
                              STATIC_TAGS.values())
elif PYTHON_VERSION.major == 3:
    BUILT_IN_TAG_VALUES = set(list(BUILT_IN_TAGS.values()) +
                              list(I18N_TAGS.values()) +
                              list(L10N_TAGS.values()) +
                              list(CACHE_TAGS.values()) +
                              list(STATIC_TAGS.values()))



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import sys

from django.core.management import call_command
from django.test import TestCase

PYTHON_VERSION = sys.version_info

if PYTHON_VERSION.major == 2:
    from StringIO import StringIO
elif PYTHON_VERSION.major == 3:
    from io import StringIO


class TestCommand(TestCase):

    def test_find_unnecessary_loads(self):
        output = StringIO()
        call_command('find_unnecessary_loads', stdout=output)
        self.assertEqual('Has issues: True', output.getvalue().strip())

        output = StringIO()
        call_command('find_unnecessary_loads', app='app', stdout=output)
        self.assertEqual('Has issues: True', output.getvalue().strip())

        output = StringIO()
        call_command('find_unnecessary_loads', app='clean', stdout=output)
        self.assertEqual('Has issues: False', output.getvalue().strip())

        output = StringIO()
        call_command('find_unnecessary_loads', app='empty', stdout=output)
        self.assertEqual('Has issues: False', output.getvalue().strip())



#!/usr/bin/env python

import os
from setuptools import find_packages, setup

setup(
    name='django-unload',
    version='0.3.1',
    url="https://github.com/Styria-Digital/django-unload",
    author='Styria Digital Services',
    description='Remove unused custom Django template tags and filters',
    long_description=open(
            os.path.join(os.path.dirname(__file__), 'README.rst')
    ).read(),
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    keywords='django template tag filter',
    install_requires=[
        'django>=1.8, <1.9',
        'tabulate==0.7.5',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)
